from .imports import *
from .tableManager import *
def get_safe_password(password):
    safe_password = quote_plus(password)
    return safe_password
# Existing utility functions remain the same
def get_dbType(dbType=None):
    return dbType or 'database'

def get_dbName(dbName=None):
    return dbName or 'abstract'

def get_dbUser(dbUser=None):
    return dbUser

def verify_env_path(env_path=None):
    return env_path or get_env_path()
def get_db_vars_from_kwargs(**kwargs):
    """
    Normalize DB-related kwargs into canonical connection variables.

    Accepted aliases (case-insensitive):
        port        -> port
        password    -> password, pass
        dbType      -> type, dbtype
        host        -> host, url, address
        dbUser      -> user, dbuser
        dbName      -> dbname, database, name
    """
    resolved = {}

    key_aliases = {
        "port": ["port"],
        "password": ["password", "pass"],
        "dbType": ["type", "dbtype"],
        "host": ["host", "url", "address"],
        "user": ["user", "dbuser"],
        "dbname": ["dbname", "database", "name"],
        "env_path":["path","env_path"]
    }

    # Normalize incoming kwargs once
    lowered_kwargs = {k.lower(): v for k, v in kwargs.items()}

    for canonical_key, aliases in key_aliases.items():
        for alias in aliases:
            if alias in lowered_kwargs:
                resolved[canonical_key] = lowered_kwargs[alias]
                break  # stop searching aliases, not keys

    return resolved
def get_kwargs_dict(**kwargs):
    return kwargs
def get_db_env_value(dbname,user,env_path=None,**kwargs):
    env_path = verify_env_path(env_path)
    dbname_part_key=""
    user_part_key=""
    if dbname:
        dbname_part_key=f"{dbname}_"
    if user:
        user_part_key=f"{dbname_part_key}{user}_"
    for key,value in kwargs.items():
        value = get_env_value(value) or value
        if not value:
            for part_key in [dbname_part_key,user_part_key]:
                temp_key = f"{part_key}{key}"
                value = get_env_value(temp_key.upper(),path=env_path)
                if value:
                    break
        kwargs[key]=value
    return get_kwargs_dict(dbname=dbname,user=user,**kwargs)
def get_db_env_key(dbname=None,user=None,port=None,password=None,host=None,env_path=None,dbType=None):
    return get_db_env_value(dbname=dbname,user=user,port=port,password=password,host=host,env_path=env_path)
def derive_db_vars(**kwargs):
    db_vars = get_db_vars_from_kwargs(**kwargs)
    return get_db_env_key(**db_vars)
def get_db_vars(**kwargs):
    dbVars = derive_db_vars(**kwargs)
    protocol = 'postgresql'
    if 'rabbit' in str(dbVars.get('dbname',"")).lower():
        protocol = 'amqp'
    dbVars['dburl'] = f"{protocol}://{dbVars['user']}:{dbVars['password']}@{dbVars['host']}:{dbVars['port']}/{dbVars['dbname']}"
    return dbVars
class connectionManager(metaclass=SingletonMeta):
        
    def __init__(self, tables=[], tables_path=None,**kwargs):
        if not hasattr(self, 'initialized'):
            self.initialized=True
            dbVars = derive_db_vars(**kwargs)
            self.env_path = dbVars.get('env_path')
            self.dbName = dbVars.get('dbname')
            self.dbType = dbVars.get('dbtype')
            self.dbUser = dbVars.get('user')
            self.dbVars = self.get_db_vars(**dbVars)
            self.user = self.dbUser = self.dbVars['user']
            self.password = self.dbVars['password']
            self.host = self.dbVars['host']
            self.port = self.dbVars['port']
            self.dbname = self.dbVars['dbname']
            self.dburl = self.dbVars['dburl']  # URL-based connection string
            self.table_mgr = TableManager()
            self.tables = tables or safe_load_from_json(file_path=tables_path) or []
            self.table_mgr.env_path = self.env_path
            self.add_insert_list=None
            
            self.check_conn()
        
    def check_conn(self):
        if self.add_insert_list == None:
##          try:
                self.table_mgr.add_insert_list(self.connect_db(), self.tables, self.dbName)
                self.table_mgr.add_insert_list(self.connect_db(), self.tables, self.dbName)
                self.add_insert_list=True
##          except:
##            pass
        return self.add_insert_list
    def get_dbName(self, dbName=None):
        return get_dbName(dbName=dbName or self.dbName)
    def get_dbType(self, dbType=None):
        return get_dbType(dbType=dbType or self.dbType)
    def get_dbUser(self, dbUser=None):
        return get_dbUser(dbUser=dbUser or self.dbUser)
    def get_env_path(self, env_path=None):
        return verify_env_path(env_path=env_path)

    def get_db_vars(self,**kwargs):
        return get_db_vars(**kwargs)

    def change_db_vars(self, tables=[], tables_path=None,**kwargs):
        dbVars = derive_db_vars(**kwargs)
        self.env_path = dbVars.get('env_path')
        self.dbName = dbVars.get('dbname')
        self.dbType = dbVars.get('dbtype')
        self.dbUser = dbVars.get('user')
        self.dbVars = self.get_db_vars(**dbVars)
        self.user = self.dbUser = self.dbVars['user']
        self.password = self.dbVars['password']
        self.host = self.dbVars['host']
        self.port = self.dbVars['port']
        self.dbname = self.dbVars['dbname']
        self.dburl = self.dbVars['dburl']  # URL-based connection string
        self.simple_connect = self.simple_connect_db()
        self.get_db_connection(self.connect_db())
        self.tables = tables or self.tables
        self.table_mgr.add_insert_list(self.connect_db(), self.tables, self.dbName)
        return self.dbVars

    def connect_db(self):
            
            """ Establish a connection to the database, either by connection parameters or via URL """
            if self.dburl:
                
                return psycopg2.connect(self.dburl)
            else:
                return psycopg2.connect(user=self.user,
                                        password=self.password,
                                        host=self.host,
                                        port=self.port,
                                        dbname=self.dbname)

    def simple_connect_db(self):
        """ Create a connection pool using the database URL """
        if self.dburl:
            return psycopg2.pool.SimpleConnectionPool(1, 10, self.dburl)
        else:
            return psycopg2.pool.SimpleConnectionPool(1, 10, user=self.user,
                                                      password=self.password,
                                                      host=self.host,
                                                      port=self.port,
                                                      database=self.dbname)

    def put_db_connection(self, conn):
        conn = conn or self.connect_db()
        self.putconn(conn)

    def get_db_connection(self):
        return self.connect_db()

    def get_insert(self, tableName):
        return self.table_mgr.get_insert(tableName)

    def fetchFromDb(self, tableName, searchValue):
        return self.table_mgr.fetchFromDb(tableName, searchValue, self.connect_db())

    def insertIntoDb(self, tableName, searchValue, insertValue):
        return self.table_mgr.insert_intoDb(tableName, searchValue, insertValue, self.connect_db())

    def search_multiple_fields(self, query, **kwargs):
        return self.table_mgr.search_multiple_fields(query=query, conn=self.connect_db())

    def get_first_row_as_dict(self, tableName=None, rowNum=1):
        return self.table_mgr.get_first_row_as_dict(tableName=tableName, rowNum=rowNum, conn=self.connect_db())
def create_connection(**kwargs):
    return connectionManager(**kwargs)

def get_db_connection(**kwargs):

    return connectionManager(**kwargs).get_db_connection()

def put_db_connection(conn):
    connectionManager().put_db_connection(conn)

def connect_db(**kwargs):
    return connectionManager(**kwargs).connect_db()

def get_insert(tableName,**kwargs):
    return connectionManager(**kwargs).get_insert(tableName)

def fetchFromDb(tableName, searchValue,**kwargs):
    return connectionManager(**kwargs).fetchFromDb(tableName, searchValue)

def insertIntoDb(tableName, searchValue, insertValue,**kwargs):
    return connectionManager(**kwargs).insertIntoDb(tableName, searchValue, insertValue)

def search_multiple_fields(query, **kwargs):
    return connectionManager().search_multiple_fields(query, **kwargs)

def get_first_row_as_dict(tableName=None, rowNum=1):
    return connectionManager().get_first_row_as_dict(tableName, rowNum)
def get_cur_conn(use_dict_cursor=True):
    """
    Get a database connection and a RealDictCursor.
    Returns:
        tuple: (cursor, connection)
    """
    conn = connectionManager().get_db_connection()
    cur = conn.cursor(cursor_factory=RealDictCursor) if use_dict_cursor else conn.cursor()
    return cur, conn



from .imports import sql,build_return_clause,build_where_clause
from .query import query_data
def update_any_combo(*,
                     table_name: str,
                     update_map: dict,
                     search_map: dict = None,
                     any_value: bool = False,
                     returning=False,        # False | True | 'count' | 'col1,col2'
                     zipit=True,
                     schema='public'):
    if not table_name or table_name == '*':
        raise ValueError("table_name is required")
    if not update_map:
        raise ValueError("update_map cannot be empty")
    search_map = search_map or {}
    base = sql.SQL("UPDATE {}.{} SET ").format(
        sql.Identifier(schema),
        sql.Identifier(table_name)
        )
    set_sql, values = build_set_clause(update_map)
    where_sql, values = build_where_clause(
        search_map,
        vals=values,
        any_value=any_value
        )
    return_sql = build_return_clause(returning=returning)
    qry = (
        base
        + set_sql
        + where_sql
        + return_sql
    )
    return query_data(qry, values=values, zipRows=zipit)



from ..init_imports import *
def require_env(key: str,env_path=None, fallback: str | None = None) -> str:
    val = get_env_value(key=key,path=env_path) or fallback
    if not val:
        raise RuntimeError(f"❌ Missing env var: {key}")
    return val



from .imports import *
def get_db_name(dbName=None, dbPath=None, dbUrl=None):
    if dbName:
        return dbName
    fileParts_js = get_file_parts(dbPath or dbUrl)
    return fileParts_js.get("fileName")

def get_db_url(dbPath=None):
    return connectionManager().dburl

def get_db_engine(dbUrl=None, dbPath=None):
    if dbUrl is None:
        if dbPath is None:
            return
        dbUrl = get_db_url(dbPath)
    return create_engine(dbUrl)

def get_all_key_values(data, parent_key='', sep='_'):
    items = {}
    for key, value in data.items():
        new_key = f"{parent_key}{sep}{key}" if parent_key else key
        if isinstance(value, dict):
            items.update(get_all_key_values(value, new_key, sep=sep))
        else:
            items[new_key] = value
    return items

def get_type_change_list():
    return ["String", "Integer", "Float", "JSON"]

def create_class_from_dict(class_name, class_dict, base):
    attributes = {
        '__tablename__': class_name.lower(),
        'id': Column(Integer, primary_key=True, autoincrement=True)
    }
    unique_keys = class_dict.get('unique_keys', [])

    def handle_nested(col_name, col_type):
        if isinstance(col_type, dict):
            return Column(JSON)
        elif isinstance(col_type, list):
            return Column(JSON)
        else:
            raise ValueError(f"Unsupported column type: {col_type}")

    for col_name, col_type in class_dict['valueKeys'].items():
        if col_type == "TEXT":
            col = Column(String)
        elif col_type == "INTEGER":
            col = Column(Integer)
        elif col_type == "FLOAT" or col_type == "REAL":
            col = Column(Float)
        elif col_type == "JSON" or col_type == "ARRAY":
            col = Column(JSON)
        elif col_type == "BOOL":
            col = Column(Boolean)
        elif isinstance(col_type, dict) or isinstance(col_type, list):
            col = handle_nested(col_name, col_type)
        else:
            raise ValueError(f"Unsupported column type: {col_type}")

        attributes[col_name] = col

        if col_name in unique_keys:
            attributes[f'ix_{col_name}'] = Index(f'ix_{class_name.lower()}_{col_name}', col, unique=True)

    return type(class_name, (base,), attributes)

# Generic function to check if a record exists
def record_exists(session, model, **kwargs):
    query = session.query(model)
    for key, value in kwargs.items():
        query = query.filter(getattr(model, key) == value)
    return query.first() is not None
class DatabaseBrowser:
    def __init__(self, dbUrl=None, dbPath=None):
        self.dbUrl = dbUrl or get_db_url(dbPath)
        self.engine = create_engine(self.dbUrl)
        self.Session = sessionmaker(bind=self.engine)
        self.session = self.Session()
        self.metadata = MetaData()
        self.inspector = sa_inspect(self.engine)
    def list_tables(self):
        tables = self.inspector.get_table_names()
        return enumerate_selections(obj_list=tables,parentObject="table",childObject=None)

    def table_list(self):
        obj_list = self.list_tables()
        return list_objects(obj_list=obj_list,object_type='table')
    def list_columns(self, table_name):
        try:
            table = Table(table_name, self.metadata, autoload_with=self.engine)
            columns = [column.name for column in table.columns]
            return enumerate_selections(obj_list=columns,parentObject="column",childObject=table)
        except Exception as e:
            print(f"Error loading table {table_name}: {e}")
            return []

    def column_list(self, table_name=None):
        if table_name is None:
            table_name = self.table_list()
        if table_name is None:
            print("No table selected.")
            return None
        obj_list = self.list_columns(table_name)  
        return list_objects(obj_list=obj_list,parentObject=table_name,object_type='column')
    def view_table(self, table_name, start=0, end=5):
        try:
            table = Table(table_name, self.metadata, autoload_with=self.engine)
            query = table.select().offset(start).limit(end - start)
            result = self.session.execute(query)
            rows = result.fetchall()

            if rows:
                df = pd.DataFrame(rows)
                df.columns = result.keys()
                pd.set_option('display.max_rows', None)
                pd.set_option('display.max_columns', None)
                pd.set_option('display.width', None)
                pd.set_option('display.max_colwidth', None)
                pd.reset_option('display.max_rows')
                pd.reset_option('display.max_columns')
                pd.reset_option('display.width')
                pd.reset_option('display.max_colwidth')
            else:
                print(f"No data found in table {table_name} from row {start} to {end}")
        except Exception as e:
            print(f"Error viewing table {table_name}: {e}")
    def alter_column_type(self, table_name, column_name, new_type):
        """Alter the type of a specific column in a table."""
        if new_type not in ['String', 'Integer', 'Float']:
            print("Invalid type. Please choose from 'String', 'Integer', or 'Float'.")
            return
        
        try:
            # Load the table
            table = Table(table_name, self.metadata, autoload_with=self.engine)
            old_column = table.c[column_name]
            
            # Determine the new column type
            if new_type == 'String':
                new_column = Column(column_name, String, nullable=old_column.nullable)
            elif new_type == 'Integer':
                new_column = Column(column_name, Integer, nullable=old_column.nullable)
            elif new_type == 'Float':
                new_column = Column(column_name, Float, nullable=old_column.nullable)
            
            # Perform the column type change
            with self.engine.connect() as connection:
                connection.execute(text(f"ALTER TABLE {table_name} RENAME COLUMN {column_name} TO {column_name}_old"))
                connection.execute(text(f"ALTER TABLE {table_name} ADD COLUMN {column_name} {new_type}"))
                connection.execute(text(f"UPDATE {table_name} SET {column_name} = {column_name}_old"))
                connection.execute(text(f"ALTER TABLE {table_name} DROP COLUMN {column_name}_old"))
                connection.commit()
            
            print(f"Column {column_name} in table {table_name} successfully altered to {new_type}.")
        except Exception as e:
            print(f"Error altering column type: {e}")

    def update_all_entries(self, table_name, column_name, new_value):
        """Update all entries in a specific column with a new value."""
        try:
            table = Table(table_name, self.metadata, autoload_with=self.engine)
            query = table.update().values({column_name: new_value})
            result = self.session.execute(query)
            self.session.commit()
            print(f"All entries in column {column_name} of table {table_name} updated to {new_value}.")
        except Exception as e:
            print(f"Error updating entries: {e}")

    def export_data_by_key_value(self, table_name,value, file_path,key='zipcode' ):
        """Export data from a specific zipcode to an Excel file."""
        try:
            query = text(f"SELECT * FROM {table_name} WHERE {key} = :{key}")
            result = self.session.execute(query, {f"{key}": value})
            rows = result.fetchall()

            if rows:
                df = pd.DataFrame(rows)
                df.columns = result.keys()
                df.to_excel(file_path, index=False)
                print(f"Data for {key} {value} exported to {file_path}")
            else:
                print(f"No data found for {key} {value} in table {table_name}")
        except Exception as e:
            print(f"Error exporting data: {e}")
    
    def search_table(self, table_name, column_name, search_value, print_value=False):
        """
        Search a table for specific values in a column.

        Args:
            table_name (str): The name of the table to search.
            column_name (str): The name of the column to search within.
            search_value (str): The value to search for in the specified column.
            print_value (bool): Whether to print the resulting DataFrame.

        Returns:
            pd.DataFrame: DataFrame containing the search results or None if no results found.
        """
        try:
            table = Table(table_name, self.metadata, autoload_with=self.engine)
        except Exception as e:
            print(f"Error loading table {table_name}: {e}")
            return
        
        # Construct the WHERE clause
        where_clause = f"{column_name} LIKE :{column_name}"
        
        try:
            query = text(f"SELECT * FROM {table_name} WHERE {where_clause}")
            result = self.session.execute(query, {column_name: f"%{search_value}%"})
            rows = result.fetchall()

            if rows:
                df = pd.DataFrame(rows)
                df.columns = result.keys()
                if print_value:
                    print(df)
                return df
            else:
                print(f"No results found for {search_value} in {column_name} of {table_name}")
        except Exception as e:
            print(f"Error executing search query: {e}")
    def search_by_json_field(self, table_name, json_field, key, value, file_path=None, case_sensitive=True, partial_match=False, save_to_excel=True):
        try:
            if not case_sensitive:
                value = value.lower()
                query = text(f"SELECT * FROM {table_name} WHERE LOWER(JSON_EXTRACT({json_field}, '$.{key}')) {'LIKE' if partial_match else '='} :val")
            else:
                query = text(f"SELECT * FROM {table_name} WHERE JSON_EXTRACT({json_field}, '$.{key}') {'LIKE' if partial_match else '='} :val")
            
            if partial_match:
                value = f"%{value}%"
            
            result = self.session.execute(query, {"val": value})
            rows = result.fetchall()
            
            if rows:
                input(rows)
                df = pd.DataFrame(rows)
                df.columns = result.keys()
                
                # Flatten the nested JSON fields
                for column in df.columns:
                    if isinstance(df[column].iloc[0], dict):
                        expanded_data = df[column].apply(lambda x: get_all_key_values(x))
                        expanded_df = pd.DataFrame(expanded_data.tolist())
                        expanded_df.columns = [f"{column}.{sub_col}" for sub_col in expanded_df.columns]
                        df = df.drop(columns=[column]).join(expanded_df)
                
                if save_to_excel and file_path:
                    df.to_excel(file_path, index=False)
                    print(f"Data for {key} {value} exported to {file_path}")
                
                return df
            else:
                print(f"No results found for {value} in {key} of {json_field} in {table_name}")
        except Exception as e:
            print(f"Error executing search query: {e}")
    def scrub_tables(self, table1_name, table2_name, join_key):
        """
        Compare and clean data between two tables based on a join key.
        
        Args:
            table1_name: Name of the first table.
            table2_name: Name of the second table.
            join_key: Key used to join the tables.
            
        Returns:
            None
        """
        table1 = Table(table1_name, self.metadata, autoload_with=self.engine)
        table2 = Table(table2_name, self.metadata, autoload_with=self.engine)

        # Select records from table1 that do not exist in table2 based on the join key
        subquery = self.session.query(table2.c[join_key]).subquery()
        query = self.session.query(table1).filter(~table1.c[join_key].in_(subquery))

        records_to_delete = query.all()
        
        # Delete records from table1 that do not exist in table2
        if records_to_delete:
            for record in records_to_delete:
                self.session.delete(record)
            self.session.commit()
            print(f"Deleted {len(records_to_delete)} records from {table1_name} not found in {table2_name}.")
        else:
            print(f"No records to delete from {table1_name}. All records match {table2_name}.")

    def scrub_clients_against_addresses(self, clients_table_name, addresses_table_name):
       
        """
        Remove records from the clients table where the corresponding address is a multi-unit dwelling with solar.
        
        Args:
            clients_table_name: Name of the clients table.
            addresses_table_name: Name of the addresses table.
            
        Returns:
            None
        """
        addresses = Table(addresses_table_name, self.metadata, autoload_with=self.engine)
        clients = Table(clients_table_name, self.metadata, autoload_with=self.engine)

        # Subquery to find addresses that are multi-unit and have solar
        subquery = self.session.query(addresses.c.address).filter(
            addresses.c.multi_unit == False,
        ).subquery()

        # Select records from clients where the address matches the subquery
        query = self.session.query(clients).filter(clients.c.address.in_(subquery))

        records_to_delete = query.all()
        
        # Delete records from clients that match the subquery
        if records_to_delete:
            for record in records_to_delete:
                self.session.delete(record)
            self.session.commit()
            print(f"Deleted {len(records_to_delete)} records from {clients_table_name} where the address is a multi-unit dwelling with solar.")
        else:
            print(f"No records to delete from {clients_table_name}. All addresses are either not multi-unit or do not have solar.")
    def get_meta_datas(self):
        obj_list = ["id","baseMint","quoteMint","tokenName","tokenSymbol"]
        field_choice = get_field_choice(obj_list=obj_list,parentObject="JSON search field",object_type='JSON field')
        value = enter_value_text(typ="the",action="to filter by",parentObject=field_choice)
        return 'pool_keys' if field_choice in ['id', 'baseMint', 'quoteMint'] else 'meta_data'
    def get_json_search_vars():
        table_name = self.table_list()
        obj_list =  self.column_list()
        field_choice = get_field_choice(obj_list=obj_list,parentObject="JSON search field",object_type='JSON field')
        value = enter_value_text(typ="the",action="to filter by",parentObject=field_choice)
        queries = {"save_to_excel":{"default":"y","query":"Save to excel?"},"partial_match":{"default":"y","query":"Partial match?"},"case_sensitive":{"default":'n',"query":"Case sensitive?"}}
        for key,values in queries.items():
            queries[key]=get_user_input(values.get("query"),default=values.get('default'))
        if save_to_excel:
          output_dir = os.path.join(os.getcwd(),'output_data')
          os.makedirs(output_dir,exist_ok=True)
          fileName = f"{table_name}_{field_choice}"
          get_user_input(values.get("query"),default=values.get('default'))
          queries = {"fileName":{"query":"please enter a fileName:","default":fileName}}
          for key,values in queries.items():
              queries[key]=get_user_input(values.get("query"),default=values.get('default'))
          file_path = os.path.join(output_dir,f"{queries['fileName']}.xlsx")
        self.search_by_json_field(table_name=table_name, json_field=json_field, key=field_choice, value=value,file_path=file_path, case_sensitive=queries["case_sensitive"], partial_match=queries["partial_match"],save_to_excel=queries["save_to_excel"])
    def export_table_to_excel(self, table_name, filters=None, file_path="output.xlsx"):
        """
        Export a table to an Excel file with optional filters.
        
        Args:
            table_name (str): Name of the table to export.
            filters (dict): Optional dictionary of column filters {column_name: value}.
            file_path (str): Path to the Excel file to save.
        """
        try:
            table = Table(table_name, self.metadata, autoload_with=self.engine)
            
            query = table.select()
            if filters:
                for column, value in filters.items():
                    query = query.where(getattr(table.c, column) == value)
                    
            result = self.session.execute(query)
            rows = result.fetchall()
            excel = []
            if rows:
                for i,row in enumerate(rows):
                    if not isinstance(row,list):
                        row = list(row)
                    if isinstance(row,tuple):
                        row = list(row)
             
        
                    excel.append(flatten_json(row, parent_key='', sep='_'))
                safe_excel_save(pd.DataFrame(excel),file_path)
                print(f"Data exported to {file_path}")
            else:
                print(f"No data found in table {table_name} with given filters.")
        except Exception as e:
            print(f"Error exporting table {table_name} to Excel: {e}")

    def get_filters(self, table_name):
        """
        Get filters from the user for a specific table.
        
        Args:
            table_name (str): Name of the table to filter.
            
        Returns:
            dict: Dictionary of filters {column_name: value}.
        """
        filters = {}
        while True:
            column_name = self.column_list(table_name)
            if not column_name:
                break
            value = input(f"Enter value to filter by {column_name} (or press Enter to skip): ")
            if value:
                filters[column_name] = value
            else:
                break
        return filters

    def delete_rows_by_criteria(self, table_name, column_name, error_code=None, error_message=None):
        """
        Delete rows from the table where the specified JSON column contains an error code or message.

        Args:
            table_name (str): The name of the table to delete rows from.
            column_name (str): The name of the column containing JSON data to check.
            error_code (int, optional): The error code to match for deletion.
            error_message (str, optional): The error message to match for deletion.

        Returns:
            None
        """
        try:
            table = Table(table_name, self.metadata, autoload_with=self.engine)
            
            # Building the query based on provided criteria
            query = table.delete().where(
                (table.c[column_name].cast(JSON)["error"]["code"].as_integer() == error_code) |
                (table.c[column_name].cast(JSON)["error"]["message"].astext == error_message)
            )
            
            result = self.session.execute(query)
            self.session.commit()
            print(f"Deleted {result.rowcount} rows from {table_name} where {column_name} contained the specified error.")
        except Exception as e:
            print(f"Error deleting rows from {table_name}: {e}")
    def execute_raw_query(self, raw_query, params=None):
        """
        Execute a raw SQL query.

        Args:
            raw_query (str): The raw SQL query string to execute.
            params (dict, optional): Parameters to pass with the query.
            print_value (bool, optional): Whether to print the result. Default is False.

        Returns:
            pd.DataFrame: DataFrame containing the query results or None if no results found.
        """
        try:
            result = self.session.execute(query, params or {})
            rows = result.fetchall()
            if rows:
                df = pd.DataFrame(rows)
                return df
            else:
                print("No results found.")
        except Exception as e:
            print(f"Error executing raw query: {e}")
            return None
    def delete_table(self, table_name: str) -> None:
        """
        Delete the specified table from the database.
        
        Args:
            table_name (str): The name of the table to delete.
        
        Returns:
            None
        """
        try:
            query = text(f"DROP TABLE IF EXISTS {table_name}")
            self.session.execute(query)
            self.session.commit()
            print(f"Table '{table_name}' has been deleted.")
        except Exception as e:
            print(f"Error deleting table '{table_name}': {e}")
    def main(self):
        while True:
            print("\nMenu:")
            print("0. Exit")
            print("1. List tables")
            print("2. Search table")
            print("3. View table contents")
            print("4. List columns in a table")
            print("5. Alter column type")
            print("6. Update all entries in a column")
            print("7. Export data by key")
            print("8. Scrub tables")
            print("9. Scrub clients against addresses")
            print("10. Export table to Excel with filters")
            print("11. Delete rows with errored transaction data")
            print("12. Delete an entire table")  # New option to delete the table
            
            choice = input("Enter your choice: ")

            if choice == "0":
                print("Exiting...")
                break
            elif choice == "1":
                self.list_tables()
            elif choice == "2":
                table_name = self.table_list()
                column_name = self.column_list(table_name)
                search_value = enter_value_text(typ="", action="to search for", parentObject=column_name)
                self.search_table(table_name, column_name, search_value)
            elif choice == "3":
                table_name = self.table_list()
                table_row_count = self.session.execute(text(f"SELECT COUNT(*) FROM {table_name}")).scalar()
                print(f"there are {table_row_count} selections in this table")
                if table_row_count:
                    start = get_integer_input(f"Enter start row (0-{table_row_count - 1})", 0, table_row_count - 1)
                    end = get_integer_input(f"Enter end row ({start + 1}-{table_row_count})", start + 1, table_row_count)
                    self.view_table(table_name, start, end)
            elif choice == "4":
                table_name = self.table_list()
                self.list_columns(table_name)
            elif choice == "5":
                column_name = self.column_list()
                field_choice = get_field_choice(obj_list=get_type_change_list(), parentObject="new type", object_type='type')
                self.alter_column_type(table_name, column_name, field_choice)
            elif choice == "6":
                column_name = self.column_list()
                new_value = enter_value_text(typ="new", action="for all entries", parentObject=column_name)
                self.update_all_entries(table_name, column_name, new_value)
            elif choice == "7":
                self.get_json_search_vars()
            elif choice == "8":
                table1_name, table2_name = self.table_list(), self.table_list()
                join_key = enter_value_text(typ="join key", action="to scrub tables by", parentObject=f"{table1_name} and {table2_name}")
                self.scrub_tables(table1_name, table2_name, join_key)
            elif choice == "9":
                clients_table_name = self.table_list()
                addresses_table_name = self.table_list()
                self.scrub_clients_against_addresses(clients_table_name, addresses_table_name)
            elif choice == "10":
                table_name = self.table_list()
                filters = self.get_filters(table_name)
                file_path = input("Enter the file path to save the Excel file (default: output.xlsx): ") or "output.xlsx"
                self.export_table_to_excel(table_name, filters, file_path)
            elif choice == "11":
                table_name = self.table_list()
                column_name = self.column_list(table_name)
                error_code = int(input("Enter the error code to search for (e.g., -32601): "))
                error_message = input("Enter the error message to search for (e.g., 'Method not found'): ")
                self.delete_rows_by_criteria(table_name, column_name, error_code, error_message)
            elif choice == "12":  # New choice to delete the entire table
                table_name = self.table_list()
                confirm = input(f"Are you sure you want to delete the entire table '{table_name}'? (y/n): ").lower()
                if confirm == 'y':
                    self.delete_table(table_name)
                else:
                    print("Table deletion canceled.")
        else:
            print("Invalid choice. Please try again.")

def manageDb(dbUrl=None,dbPath=None):
    db_browser = DatabaseBrowser(dbUrl=dbUrl,dbPath=dbPath)
    db_browser.main()



from .imports import get_db_from,fetch_any_combo
from .logdata import get_signature_from_log_id
def get_transaction_from(columnNames=None,searchColumn=None,searchValue=None,count=False,zipit=True):
    columnNames = columnNames or '*'
    return get_db_from(tableName='transactions',
                       columnNames=columnNames,
                       searchColumn=searchColumn,
                       searchValue=searchValue,
                       count=count,
                       zipit=zipit)
def count_transactions_for_pair(pair_id,zipit=True):
    return get_transaction_from(searchColumn="pair_id",searchValue=pair_id,count=True,zipit=zipit)
def get_transaction_from_signature(signature,zipit=True):
    return get_transaction_from(searchColumn="signature",searchValue=signature,zipit=zipit)
def get_transaction_from_txn_id(txn_id,zipit=True):
    return get_transaction_from(searchColumn="id",searchValue=txn_id,zipit=zipit)
def get_transaction_from_log_id(log_id,zipit=True):
    signature = get_signature_from_log_id(log_id,zipit=zipit)
    return get_transaction_from_signature(signature,zipit=zipit)
def get_all_txns_from_pair_id(pair_id,zipit=True):
    return get_transaction_from(searchColumn="pair_id",searchValue=pair_id,zipit=zipit)
def get_transactions(txn_id=None,log_id=None,signature=None,pair_id=None,zipit=True):
    if pair_id:
        return get_all_txns_from_pair_id(pair_id,zipit=zipit)
    if txn_id:
        return get_transaction_from_txn_id(txn_id,zipit=zipit)
    if log_id:
        return get_transaction_from_log_id(log_id,zipit=zipit)
    if signature:
        return get_transaction_from_signature(signature,zipit=zipit)




from .imports import *
def get_rows(rows):
    if not rows:
        return None
    if isinstance(rows,psycopg2.extras.RealDictRow):
        rows = dict(rows)
    if isinstance(rows, list):
        for i,row in  enumerate(rows):
            if isinstance(row,psycopg2.extras.RealDictRow):
                row = dict(row)
            rows[i] = row
    # If select_rows returned a dict, use it; if it returned a list, grab the first item
    if isinstance(rows, dict):
        return rows
    else:
        return rows
def get_last_row(table_name,**kwargs):
    dbName = kwargs.get('dbName','solcatcher')
    with get_engine(dbName).connect() as conn:
        result = conn.execute(text(f"SELECT * FROM {table_name} ORDER BY id DESC LIMIT 1;"))
        last_row = result.fetchone()
        return last_row



from .imports import *
class TableManager(metaclass=SingletonMeta):
    def __init__(self):
        if not hasattr(self, 'initialized'):
            self.initialized = True
            self.insert_list = []
            self.dbNames=[]
            self.env_path = None
    def add_insert_list(self,conn,tables,dbName):
        self.dbNames.append(dbName)
        setup_database(tables,conn)
        self.insert_list+=tables
    def get_insert(self, tableName):
        tableName = tableName.lower()
        insertList = [ls for ls in self.insert_list if ls.get("tableName") == tableName.lower()]
        return insertList[0] if insertList else None

    def fetchFromDb(self,tableName,searchValue,conn):
        cached_response = fetchFromDb(self.get_insert(tableName), searchValue, conn)
        if cached_response:
            return cached_response

    def insert_intoDb(self,tableName,searchValue,insertValue,conn):
        return insert_intoDb(self.get_insert(tableName), searchValue,insertValue,conn )
    def search_multiple_fields(self, query,conn,**kwargs):
        return search_multiple_fields(query,conn)
    def get_first_row_as_dict(self,tableName=None,conn=None,rowNum=None):
        """Fetch the first row of data from the specified table and return as a dictionary."""
        rowNum= rowNum or 1
        query = f"SELECT * FROM {tableName} ORDER BY id ASC LIMIT {rowNum};"
        cur = conn.cursor()
        try:
            cur.execute(query)
            first_row = cur.fetchone()
            col_names = [desc[0] for desc in cur.description]
            if first_row:
                return dict(zip(col_names, first_row))
            return None
        except psycopg2.Error as e:
            print(f"Error fetching the first row: {e}")
            return None
        finally:
            cur.close()
            conn.close()
    def add_table_config(self,dbName,dbType):
        file_path = get_table_path(dbName,dbType)
        logger.info(file_path)
        if file_path:
            table_config = safe_load_from_json(file_path=file_path)
            self.insert_list+=table_config
    def add_unique_header(self,uniqueHeader,tableName,dbName,dbType):
        table_path = get_table_path(dbName,dbType)
        logger.info(table_path)
        if table_path:
            insertTables = safe_load_from_json(file_path=table_path)
            newTables = []
            tableName = tableName.lower()
            for i,insertTable in enumerate(insertTables):
                if tableName == insertTable.get('tableName'):
                    if 'excelUniqueHeaders' not in insertTable:
                        insertTable['excelUniqueHeaders'] = []
                    if uniqueHeader not in insertTable['excelUniqueHeaders']:
                        insertTable['excelUniqueHeaders'].append(uniqueHeader)
                        self.insert_list = [insertTable if table.get('tableName') == tableName else table for table in self.insert_list]
                newTables.append(insertTable)
            safe_dump_to_file(data=newTables,file_path=table_path)

def get_insert_tables(dbName,dbType,env_path = None):
    tablePath = get_table_path(dbName,dbType,env_path=env_path)
    return safe_read_from_json(tablePath)
def get_insert_table(tableName):
    insertTable = [table for table in TableManager().insert_list if tableName == table.get('tableName').lower()]
    if insertTable:
        insertTable= insertTable[0]
    return insertTable or {}
def get_insert(tableName):
    tableName = tableName.lower()
    return TableManager().get_insert(tableName)
    
def add_unique_header(uniqueHeader,tableName,dbName,dbType):
    TableManager().add_unique_header(uniqueHeader,tableName,dbName,dbType)
def add_table_config(dbName,dbType):
    TableManager().add_table_config(dbName,dbType)
def get_table_path(dbName,dbType,env_path=None):
    key = f"{dbName.upper()}_{dbType.upper()}_CONFIGPATH"
    return get_env_value(path=env_path or TableManager().env_path,key=key)


def getInsertType(tableName):
    return TableManager().get_insert(tableName)

async def search_Db(tableName, searchValue,**kwargs):
    insertType = getInsertType(tableName)
    if not insertType:
        print(f"No data structure for {tableName}")
        return

    search_query = getSearchQuery(insertType.get('tableName'), '*', insertType.get('columnSearch'))
    try:
        result = await getQuery(search_query, (searchValue,))
        if result:
            print(f"Found in {tableName}: {result}")
            return result
        else:
            print(f"No data found in {tableName} for {searchValue}")
    except Exception as e:
        print(f"Error searching in {tableName}: {str(e)}")
        
async def insert_Db(tableName, searchValue, insertValue,**kwargs):
    insertType = getInsertType(tableName)
    if not insertType:
        print(f"No insert type found for table: {tableName}")
        return

    # Check if entry exists
    existing_entry = await search_Db(insertType.get('tableName'), searchValue)
    if existing_entry:
        print(f"Entry already exists in {tableName} with value {searchValue}")
        return existing_entry

    if not isinstance(insertValue, tuple):
        insertValue = (searchValue, dump_if_json(insertValue))
    
    insert_query = "INSERT INTO {} ({}) VALUES ({})".format(insertType.get('tableName'), insertType('insertQuery'), getInsertQueryS(insertType('insertQuery')))
    try:
        await getQuery(insert_query, insertValue)
        #print(f"Inserted into {tableName}: {searchValue}")
    except Exception as e:
        print(f"Error using {insert_query} inserting {instype} with a value of {query} into {tableName}: {str(e)}")




async def universal_db_function(method,  params, check_identical_params=None, doNotCall=None, url=None, data={}):
    # Ensure the table is ready for operations
    insertType = getInsertType(method)
    if not inputs:
        print(f"Unable to find or create table: {method}")
        return

    # Extract the unique identifier for database operations
    unique_identifier = params[0]

    # Attempt to find an existing entry in the database
    existing_entry = await search_Db(tableName, unique_identifier)

    # Determine if the existing entry should be ignored based on the parameters
    if check_identical_params and existing_entry:
        all_match = all(params[k] == existing_entry.get(k, None) for k in params if k in existing_entry)
        if not all_match:
            existing_entry = None

    # If instructed not to make external calls and an entry exists, return it
    if doNotCall and existing_entry:
        return existing_entry

    # If there's no existing entry or it's ignored based on parameters, make an external call if allowed
    if not existing_entry and not doNotCall:
        response_data = await (asyncPostRequest(url=url, data=data, endpoint=method) if url else makeLimitedCall(method, params))

        # Handle unsuccessful responses or errors
        if not response_data or (isinstance(response_data, dict) and response_data.get('error')):
            print(f"Failed to get a valid response for {unique_identifier} in {tableName}: {response_data}")
            return response_data

        # If the response is successful, insert it into the database
        await insert_Db(insertType.get('tableName'), unique_identifier, (unique_identifier, dump_if_json(response_data)))
        #print(f"Inserted {unique_identifier} in {tableName}")
        return response_data

    # Return the existing entry if all checks pass and no external call is made
    return existing_entry

def getSearchQuery(tableName, valueSelect='*', columnName=''):
    return f"SELECT {valueSelect} FROM {tableName} WHERE {columnName} = $1"

def getsearchquery(tableName):
    insertType = getInsertType(tableName)
    return getSearchQuery(insertType.get('tableName'), '*', insertType['columnSearch'])

def getInsertQueryS(insertQuery):
    parts = [part.strip() for part in insertQuery[1:-1].split(',')]
    placeholders = ', '.join(f'${i + 1}' for i in range(len(parts)))
    return placeholders


def setup_database(tables, conn):
    """ Create database tables based on provided configurations """
    cur = conn.cursor()
    try:
        for table in tables:
            cur.execute(table['table'])
        conn.commit()
    except psycopg2.Error as e:
        print("Error setting up database tables:", e)
        conn.rollback()
    finally:
        cur.close()

def fetch_data(search_query, search_value, conn):
    """ Fetch data from the database using the specified query and value """
    cur = conn.cursor()
    try:
        cur.execute(search_query, (search_value,))
        return cur.fetchone()
    finally:
        cur.close()

def insert_data(insert_query, values, conn):
    """ Insert data into the database using the specified query and values """
    cur = conn.cursor()
    try:
        cur.execute(insert_query, values)
        conn.commit()
    except psycopg2.Error as e:
        print("Failed to insert data:", e)
        conn.rollback()
    finally:
        cur.close()
        
def search_multiple_fields(query,conn,**kwargs):
    """Search the database using a query."""
    cur = conn.cursor()
    try:
        cur.execute(query)
        return cur.fetchall()
    except psycopg2.Error as e:
        print(f"Error querying JSONB data: {e}")
    finally:
        cur.close()
        conn.close()
def perform_database_operations(operation_type, tableName, searchValue,insertValue, conn):
    """ Perform dynamic database operations based on type and table info """
    insertInfo = get_insert(tableName)
    if operation_type == 'fetch':
        return fetch_data(insertInfo['searchQuery'], searchValue, conn)
    elif operation_type == 'insert':
        insert_data(insertInfo['insertQuery'], searchValue, conn)

def fetchFromDb(insertInfo,searchValue,conn):
    cached_response = fetch_data(insertInfo['searchQuery'], searchValue, conn)
    if cached_response:
        return cached_response

def insert_intoDb(insertInfo,searchValue,insertValue,conn):
    if isinstance(insertValue,dict):
        insertValue = json.dumps(insertValue)
    return insert_data(insertInfo['insertQuery'], (searchValue,insertValue), conn)



# src/env/queues.py
from .imports import *

_QUEUE_DISPLAYED = False


def load_queue_env(env_path=None) -> dict:
    global _QUEUE_DISPLAYED

    out = {
        "workerName": require_env(key="SOLCATCHER_QUEUE_NAME",fallback="solcatcher",env_path=env_path),

        "vhost": require_env(key="SOLCATCHER_QUEUE_VHOST",fallback="solcatcher",env_path=env_path),
        "logIntake": require_env(key="SOLCATCHER_QUEUE_LOG_INTAKE",fallback="logIntakeQueue",env_path=env_path),
        "logEntry": require_env(key="SOLCATCHER_QUEUE_LOG_ENTRY",fallback="logEntryQueue",env_path=env_path),
        "txnEntry": require_env(key="SOLCATCHER_QUEUE_TXN_ENTRY",fallback="txnEntryQueue",env_path=env_path),

        "metadataGenesisEnrich": require_env(key=
            "SOLCATCHER_QUEUE_METADATA_GENESIS_ENRICH_QUEUE",fallback="metadataGenesisEnrichQueue"
        ,env_path=env_path),
        "metadataUriEnrich": require_env(key=
            "SOLCATCHER_QUEUE_METADATA_URI_ENRICH_QUEUE",fallback="metadataUriEnrichQueue"
        ,env_path=env_path),

        "signatureCall": require_env(key="SOLCATCHER_QUEUE_SIGNATURE_CALL",fallback="signatureCallQueue",env_path=env_path),
        "metaDataCall": require_env(key="SOLCATCHER_QUEUE_META_DATA_CALL",fallback="metaDataCallQueue",env_path=env_path),
        "pairCall": require_env(key="SOLCATCHER_QUEUE_PAIR_ENTRY",fallback="pairCallQueue",env_path=env_path),
        "signatureGenesisCall": require_env(key=
            "SOLCATCHER_QUEUE_SIGNATURE_GENESIS_CALL",fallback="signatureGenesisCallQueue"
        ,env_path=env_path),
        "rpcCall": require_env(key="SOLCATCHER_QUEUE_RPC_CALL",fallback="rpcCallQueue",env_path=env_path),
        "transactionCall": require_env(key="SOLCATCHER_QUEUE_TRANSACTION_CALL",fallback="transactionCallQueue",env_path=env_path),
        "getSignaturesCall": require_env(key=
            "SOLCATCHER_QUEUE_GET_SIGNATURES_CALL",fallback="getSignaturesCallQueue"
        ,env_path=env_path),
    }

    if not _QUEUE_DISPLAYED:
        print("📦 Queue config:", out)
        _QUEUE_DISPLAYED = True

    return out




from abstract_math import divide_it
from abstract_pandas import safe_excel_save
from abstract_utilities import *
from abstract_security import *
from abstract_apis import (
    getRequest,
    get_response,
    asyncPostRpcRequest,
    asyncPostRequest
    )











from .imports import *
# Assuming getRequest is defined in globalImports, import it explicitly


def fetch_image(uri, max_size=(200, 200)):
    """
    Fetch an image from the given URI, verify its integrity, resize it,
    and return the processed image data as PNG bytes.
    
    The function first attempts to extract an image URL from the URI via
    get_img_url. If none is found, it falls back to using the original URI.
    """
    logging.info(f"Fetching image from URI: {uri}")
    try:
        # Attempt to extract the image URL from the provided URI.
        image_url = get_img_url(uri)
        if not image_url:
            logging.error("Image URL could not be retrieved. Falling back to the provided URI.")
            image_url = uri

        # Fetch the image content.
        content = get_image_content(image_url)
        if not content:
            logging.error("Image content could not be retrieved.")
            return None

        # Load the image from the content bytes and verify integrity.
        image_data = io.BytesIO(content)
        img = Image.open(image_data)
        img.verify()  # Verify image integrity

        # Re-open the image for processing.
        image_data.seek(0)
        img = Image.open(image_data)
        img.thumbnail(max_size)  # Resize image to fit within max_size

        # Save the processed image to a bytes buffer.
        bio = io.BytesIO()
        img.save(bio, format="PNG")
        bio.seek(0)
        return bio.getvalue()

    except requests.HTTPError as http_err:
        logging.error(f"HTTP error occurred: {http_err}")
    except requests.RequestException as req_err:
        logging.error(f"Request exception: {req_err}")
    except Image.UnidentifiedImageError:
        logging.error("Failed to identify the image file.")
    except Exception as e:
        logging.error(f"Unexpected error: {e}")

    return None


def get_image_content(image_url):
    """
    Fetch content from the image URL and ensure it is an image.
    
    Raises:
        ValueError: If the Content-Type does not indicate an image.
    Returns:
        The raw bytes of the image.
    """
    try:
        # Perform a GET request to fetch the image data.
        response = requests.get(image_url)
        response.raise_for_status()

        # Check if the returned content is an image.
        content_type = response.headers.get('Content-Type', '')
        if not content_type.startswith('image/'):
            logging.error(f"URI does not point to an image. Content-Type: {content_type}")
            raise ValueError("Non-image content")

        return response.content

    except requests.HTTPError as http_err:
        logging.error(f"HTTP error occurred while fetching image content: {http_err}")
        raise
    except requests.RequestException as req_err:
        logging.error(f"Request exception while fetching image content: {req_err}")
        raise
    except Exception as e:
        logging.error(f"Unexpected error in get_image_content: {e}")
        raise


def get_img_url(uri):
    """
    Extract the image URL from the given URI by fetching image variables.
    
    The function calls get_image_vars, which is expected to return a JSON-like
    dictionary with an 'image' key. If the extraction fails, returns None.
    """
    logging.info(f"Extracting image URL from URI: {uri}")
    try:
        uri_vars = get_image_vars(uri)
        # Expecting uri_vars to be a dictionary containing the image URL.
        return uri_vars.get('image')
    except Exception as e:
        logging.error(f"Error extracting image URL from URI vars: {e}")
        return None


def get_image_vars(uri):
    """
    Fetch and parse image variables from the given URI.
    
    Assumes the response is JSON and either returns a parsed dictionary or,
    if getRequest already returns JSON, passes it through.
    """
    try:
        # Call getRequest (assumed to be a wrapper around requests.get).
        response = getRequest(url=uri, data={})
        
        # If response is already a dict, assume it is the parsed JSON.
        if isinstance(response, dict):
            return response
        else:
            # Otherwise, try to parse the response as JSON.
            return response.json()
    except requests.HTTPError as http_err:
        logging.error(f"HTTP error occurred while fetching image vars: {http_err}")
        raise
    except requests.RequestException as req_err:
        logging.error(f"Request exception while fetching image vars: {req_err}")
        raise
    except ValueError as ve:
        logging.error(f"JSON decoding failed: {ve}")
        raise
    except Exception as e:
        logging.error(f"Unexpected error in get_image_vars: {e}")
        raise


def get_image_data(image_url, max_size=(200, 200)):
    """
    Alternative function to fetch and process an image given an image URL.
    
    This function directly uses requests.get (without getRequest) to obtain
    the image, resize it, and return the image data as PNG bytes.
    """
    try:
        response = requests.get(image_url)
        response.raise_for_status()
        image = Image.open(io.BytesIO(response.content))
        image.thumbnail(max_size)
        bio = io.BytesIO()
        image.save(bio, format="PNG")
        return bio.getvalue()
    except Exception as e:
        logging.error(f"Error in get_image_data: {e}")
        return None



from .imports import *
def manage_db():
    manageDb(dbUrl=get_db_url())
def get_from_db(freshCall,tableName,searchValue,fetch_from_db):
    if not freshCall:
        insertValue  = fetch_from_db(tableName=tableName,searchValue=searchValue)
        if insertValue:
            return insertValue

def find_latest_index():
    df_mgr = dfManager()
    last_row = get_last_row("realestatedata")
    if last_row:
        search_address = last_row[1]
        for index, row in df_mgr.df.iterrows():
            if search_address == df_mgr.get_address_from_row(row):
                return index
    return 0
def query_table(dbName,*args,**kwargs) -> None:
    with get_engine(dbName).connect() as conn:
        try:
            conn.execute(*args,**kwargs)
            conn.commit()
        except SQLAlchemyError as e:
            print(f"Failed to insert data: {e}")
            return e
def get_browser_mgr():
    data_brswr = DatabaseBrowser(dbUrl=get_db_url())
    return data_brswr
def ensure_table_exists(dbName,tableName):
    table = get_table(tableName)
    query_table(dbName,get_browser_mgr(), table.get('table'))
def delete_table(tableName):
    get_browser_mgr().delete_table(tableName.replace('_','').lower())
def view_table(tableName):
    brows_mgr = get_browser_mgr()
    table_row_count = brows_mgr.session.execute(text(f"SELECT COUNT(*) FROM {tableName}")).scalar()
    if table_row_count:
        brows_mgr.view_table(tableName, table_row_count-1, table_row_count)
def insertintodefaultvalues(dbName,insertTable,key=None, value=None,tableName=None):
    #ensure_table_exists(tableName)
    key_key = insertTable.get('columnSearch')
    insertName=insertTable.get('insertName')
    tableName = insertTable.get('tableName')
    # Adjusted insert query using consistent parameter style
    insert_query = text(f"""
    INSERT INTO {tableName} ({key_key}, {insertName}, last_updated)
    VALUES (:{key_key}, :{insertName}, NOW())
    ON CONFLICT ({key_key}) DO UPDATE
    SET {insertName} = EXCLUDED.{insertName}, last_updated = NOW()
    WHERE {tableName}.{insertName} != EXCLUDED.{insertName};
    """)
    # Convert value toinput JSON string if it's not a JSON-compatible type
    json_value = Json(value) if isany_instance(value) else Json(f'{value}')
    key_value = str(key) if isany_instance(key) else str(f'{key}')
    # Connect to the database using SQLAlchemy
    e = query_table(dbName,insert_query, {key_key: key_value, insertName: json_value})
    if e is None:
        print(f"Inserted ({key}, value) as json_value successfully.")



from .imports import *
def get_conn_mgr():
    print('get_connection')
    return connectionManager(
                  dbType='abstract_base',
                  dbName='admin')
def get_cur_conn(use_dict_cursor=True):
    """
    Get a database connection and a RealDictCursor.
    Returns:
        tuple: (cursor, connection)
    """
    conn = connectionManager().get_db_connection()
    cur = conn.cursor(cursor_factory=RealDictCursor) if use_dict_cursor else conn.cursor()
    return cur, conn

def get_connection(env=None,env_path=None):
    """Establish a PostgreSQL connection."""
    env = env or load_postgres_env(env_path)
    name_keys = ['dbname','dbName','database']
    for key in name_keys:
        dbname = env.get(key)
        if key in env:
            del env[key]
    env['dbname']=dbname
    del env['url']
    return psycopg2.connect(**env)



from time import time
import setuptools
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name='abstract_database',
    version='0.0.2.120',
    author='putkoff',
    author_email='partners@abstractendeavors.com',
    description="A lightweight' , 'env-driven toolkit for Postgres connections' , 'table helpers' , 'and Pandas exports.",
    long_description=long_description,
    long_description_content_type='text/markdown',
    url="https://github.com/AbstractEndeavors/abstract_database",

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.11',
    ],
    install_requires=['pillow' , 'abstract_apis' , 'abstract_database' , 'abstract_gui' , 'abstract_math' ,'abstract_pandas' ,
                      'abstract_security' ,'abstract_utilities' ,'numpy' ,'pandas' , 'psycopg2' ,'sqlalchemy', 'abstract_pandas'],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.6",
    # Add this line to include wheel format in your distribution
    setup_requires=['wheel'],
)

 



from .imports import *
def get_assigned_account(address):
    query = "SELECT assigned_account FROM wallet_account_assignments WHERE %s = ANY(addresses);"  # Assuming 'addresses' is the correct column name that holds an array.
    assigned_account = query_data(query, (address,))  # Ensure the parameter is passed as a tuple
    if assigned_account:
        assigned_account = assigned_account[0]['assigned_account']  # Extract the first result if available
    return assigned_account
def assert_user_wallets(profits):
    account_assignments = []
    for user_address,values in profits.items():
        signature = values.get('signature')
        response = update_and_add_wallet(user_address,signature)
        account_assignments.append(response)
    return account_assignments
def get_pairs_from_user_wallets(user_wallets):
    pairs = fetch_any_combo(columnNames='*',tableName='pairs',searchColumn='user_address',searchValue=user_wallets,anyValue=True,zipit=True)
    return pairs
